package test.testSOP.algorithms.multicloud;

import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import qcg.shared.constants.BrokerConstants;

import dcworms.schedframe.scheduling.ExecTask;

import schedframe.events.scheduling.SchedulingEvent;
import schedframe.resources.ResourceStatus;
import schedframe.resources.StandardResourceType;
import schedframe.resources.computing.ComputingResource;
import schedframe.resources.computing.Node;
import schedframe.resources.computing.Processor;
import schedframe.resources.computing.profiles.energy.power.StandardPowerStateName;
import schedframe.resources.units.Memory;
import schedframe.resources.units.ProcessingElements;
import schedframe.resources.units.ResourceUnit;
import schedframe.resources.units.ResourceUnitName;
import schedframe.resources.units.StandardResourceUnitName;
import schedframe.scheduling.manager.resources.ClusterResourceManager;
import schedframe.scheduling.manager.resources.ResourceManager;
import schedframe.scheduling.manager.tasks.JobRegistry;
import schedframe.scheduling.manager.tasks.JobRegistryImpl;
import schedframe.scheduling.plan.SchedulingPlanInterface;
import schedframe.scheduling.plan.impl.SchedulingPlan;
import schedframe.scheduling.plugin.ModuleList;
import schedframe.scheduling.queue.TaskQueue;
import schedframe.scheduling.queue.TaskQueueList;
import schedframe.scheduling.tasks.TaskInterface;
import simulator.DataCenterWorkloadSimulator;
import simulator.stats.implementation.DCWormsStatistics;
import test.testSOP.ComparatorHostHighestLoadFirst;
import test.testSOP.ComparatorHostLowestLoadFirst;
import test.testSOP.Utility;
import example.localplugin.BaseLocalSchedulingPlugin;
import gridsim.GridSimTags;
import gridsim.dcworms.DCWormsTags;

public class SOPVP extends BaseLocalSchedulingPlugin {
	Log log = LogFactory.getLog(SOPVP.class); 
	
	public static String MIGRATION_PATH = "src/test/testSOP_results/data/multicloud/migrations.txt";
	public static String ON_VS_EXPECTED = "src/test/testSOP_results/data/multicloud/on_vs_expected.txt";
	public static String COMPUTATION_TIME = "src/test/testSOP_results/data/multicloud/computation_time.txt";
	public static String TASK_REJECTIONS = "src/test/testSOP_results/data/multicloud/task_rejections.txt";

	
	public SchedulingPlanInterface<?> schedule(SchedulingEvent event, TaskQueueList queues, JobRegistry jobRegistry,
			ResourceManager resManager, ModuleList modules) {

		ClusterResourceManager resourceManager = (ClusterResourceManager) resManager;
//		List<Processor> processors = resourceManager.getProcessors();
//		List<Node> nodes = resourceManager.getNodes();
		List<Node> nodes = Utility.getNodesByPowerState(resourceManager.getNodes(), StandardPowerStateName.ON);
		SchedulingPlan plan = new SchedulingPlan();
		// choose the events types to serve.
		// Different actions for different events are possible.
		switch (event.getType()) {
		case START_TASK_EXECUTION:
		case TASK_FINISHED:
			// our tasks are placed only in first queue (see BaseLocalSchedulingPlugin.placeJobsInQueues() method)
			TaskQueue q = queues.get(0);
			// check all tasks in queue

			for (int i = 0; i < q.size(); i++) {
				TaskInterface<?> task = q.get(i);
				// if status of the tasks in READY
				if (task.getStatus() == DCWormsTags.READY) {
					Map<ResourceUnitName, ResourceUnit> choosenResources = chooseResourcesForExecution(nodes, task);
					if (choosenResources != null) {
						addToSchedulingPlan(plan, task, choosenResources);
					}
					else {
						sendBackTask(task, jobRegistry, q, resourceManager.getNodes().get(0).getFullName().split("/")[0]); 
					}
					//FOR TESTING PURPOSE
//					if (resourceManager.getNodes().get(0).getFullName().contains("ResDataCenter_2")) {
//						
//						sendBackTask(task, jobRegistry, q, resourceManager.getNodes().get(0).getFullName().split("/")[0]); 
//					}
//					else {
//						Map<ResourceUnitName, ResourceUnit> choosenResources = chooseResourcesForExecution(nodes, task);
//						if (choosenResources != null) {
//							addToSchedulingPlan(plan, task, choosenResources);
//						} 
//					}
				}
			}
			break;
//		case POWER_LIMIT_EXCEEDED:
////			String src = event.getSource();
////			optimizeEnergyUsage(jobRegistry, resourceManager , src);
//			break;
			
		case TIMER:
			
			log.debug("Scheduler Invoked by timer");
			long startTime = System.nanoTime();
			consolidate(jobRegistry, resourceManager);
			long endTime = System.nanoTime();
			long duration = (endTime - startTime) / 1000000;  //divide by 1000000 to get milliseconds.
			try {
				PrintWriter writer = new PrintWriter(new FileWriter(COMPUTATION_TIME, true));
				writer.write(String.format("%d\t%d\n", resourceManager.getNodes().size(), duration));
				writer.close();
			} catch (IOException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
			
			Utility.manageHosts(jobRegistry, resourceManager, log);
		case JOB_ARRIVED:
			break;
		case JOB_CANCELED:
			break;
		case JOB_FAILED:
			break;
		case JOB_FINISHED:
			break;
		case PROVIDER_FAILED:
			break;
		case RESOURCE_FAILED:
			break;
		case RESOURCE_STATE_CHANGED:
			break;
		case TASK_ARRIVED:
			break;
		case TASK_CANCELED:
			break;
		case TASK_FAILED:
			break;
		case TASK_MOVED:
			break;
		case TASK_PAUSED:
			break;
		case TASK_REQUESTED_TIME_EXPIRED:
			break;
		case TASK_RESUMED:
			break;
		default:
			break;
		}
		
		return plan;
	}


	

	


	


	/**
	 * Consolidate
	 * @param jobRegistry
	 * @param resourceManager
	 */
	@SuppressWarnings("unchecked")
	private void consolidate(JobRegistry jobRegistry, ClusterResourceManager resourceManager) {
		boolean success = true;
		List<Node> nodeList = (List<Node>) resourceManager.getResourcesByTypeWithStatus(StandardResourceType.Node, ResourceStatus.FREE);
		Node leastLoadedNode = this.getLeastLoadedHost(nodeList);
		
		//all nodes are empty
		if (leastLoadedNode == null) {
			log.debug("All nodes are empty");
			return;
		}
		log.debug("Least loaded node is "+leastLoadedNode.getFullName());
	
		//Sort host lowest load first
		Collections.sort(nodeList, new ComparatorHostLowestLoadFirst());
		
		//Print the nodelist
//		this.printNodes(nodeList);
		
		ArrayList<Node> hostListMostCPU = new ArrayList<Node>();
		ArrayList<Node> hostListMostMEM = new ArrayList<Node>();
		ArrayList<String> migrateVM= new ArrayList<String>(); //VM to migrate
		ArrayList<String> migrateHost= new ArrayList<String>(); //Host to migrate to
		
		//Sort in 2 lists
		for (Node h : nodeList) {
			if (h == leastLoadedNode)
				continue;
			if (getCPULoad(h, null, null, null) > getMEMLoad(h, null, null, null)){
				hostListMostCPU.add(h);
			}
			else {
				hostListMostMEM.add(h);
			}
		}
		
		JobRegistry leastLoadedNodeJR = new JobRegistryImpl(leastLoadedNode.getFullName());
		List<ExecTask> runningTasks = leastLoadedNodeJR.getRunningTasks();
		//For each VM
		for (int i = 0 ;  i<runningTasks.size() && success==true ; i++) {
			ExecTask t = runningTasks.get(i);
//			VM vm = getVMFromId(runningTasks.get(i), vmList);
			boolean found = false;
//			log.debug("There are "+hostListMostCPU.size()+" hosts with More CPU used than MEM");
//			log.debug("There are "+hostListMostMEM.size()+" hosts with More MEM used than CPU");
			//if VM is CPU bound
			if (this.taskIsCPUBound(t)) {
//				log.debug("Task "+t.getJobId()+" is CPU bound");
				//Try MOSTMEMLIST first
				for (int h = 0 ; h<hostListMostMEM.size() && !found ; h++) {
					Node otherhost = hostListMostMEM.get(h);
					if (taskCanFitOnNode(t, otherhost, jobRegistry.getRunningTasks(), migrateVM, migrateHost) && otherhost != leastLoadedNode && !Utility.nodeIsEmpty(otherhost, log)) {
//						log.debug("Task "+t.getJobId()+" can fit on host in MOSTMEMLIST "+otherhost.getFullName());
						//register migration
						migrateVM.add(t.getJobId());
						migrateHost.add(otherhost.getFullName());
						//update usage
//						otherhost.setCpuUsage(otherhost.getCpuUsage()+vm.getCpuNumber()*100);
//						otherhost.setMemUsage(otherhost.getMemUsage()+vm.getMem());
						//If needed change list of h
						if (getCPULoad(otherhost, jobRegistry.getRunningTasks(), migrateVM, migrateHost) < getMEMLoad(otherhost, jobRegistry.getRunningTasks(), migrateVM, migrateHost)) {
							hostListMostMEM.remove(otherhost);
							hostListMostCPU.add(otherhost);
						}
						found = true;
					}
				}
				//Then try MOSTCPULIST
				for (int h = 0 ; h<hostListMostCPU.size() && !found ; h++) {
					Node otherhost = hostListMostCPU.get(h);
					if (taskCanFitOnNode(t, otherhost, jobRegistry.getRunningTasks(), migrateVM, migrateHost) && otherhost != leastLoadedNode && !Utility.nodeIsEmpty(otherhost, log)) {
//						log.debug("Task "+t.getJobId()+" can fit on host in MOSTCPULIST "+otherhost.getFullName());
						//register migration
						migrateVM.add(t.getJobId());
						migrateHost.add(otherhost.getFullName());
						//update usage
//						otherhost.setCpuUsage(otherhost.getCpuUsage()+vm.getCpuNumber()*100);
//						otherhost.setMemUsage(otherhost.getMemUsage()+vm.getMem());
						//If needed change list of h
						if (getCPULoad(otherhost, jobRegistry.getRunningTasks(), migrateVM, migrateHost) < getMEMLoad(otherhost, jobRegistry.getRunningTasks(), migrateVM, migrateHost)) {
							hostListMostCPU.remove(otherhost);
							hostListMostMEM.add(otherhost);
						}
						found = true;
					}
				}
			}
			//If VM is MEM bound
			else{
//				log.debug("Task "+t.getJobId()+" is MEM bound");
				//try MOSTCPULIST first
				for (int h = 0 ; h<hostListMostCPU.size() && !found ; h++) {
					Node otherhost = hostListMostCPU.get(h);
					if (taskCanFitOnNode(t, otherhost, jobRegistry.getRunningTasks(), migrateVM, migrateHost) && otherhost != leastLoadedNode && !Utility.nodeIsEmpty(otherhost, log)) {
//						log.debug("VM "+t.getJobId()+" can fit on host in MOSTCPULIST "+otherhost.getFullName());
						//register migration
						migrateVM.add(t.getJobId());
						migrateHost.add(otherhost.getFullName());
						//update usage
//						otherhost.setCpuUsage(otherhost.getCpuUsage()+vm.getCpuNumber()*100);
//						otherhost.setMemUsage(otherhost.getMemUsage()+vm.getMem());
						//If needed change list of h
						if (getCPULoad(otherhost, jobRegistry.getRunningTasks(), migrateVM, migrateHost) < getMEMLoad(otherhost, jobRegistry.getRunningTasks(), migrateVM, migrateHost)) {
							hostListMostCPU.remove(otherhost);
							hostListMostMEM.add(otherhost);
						}
						found = true;
					}
				}
				//then MOSTMEMLIST 
				for (int h = 0 ; h<hostListMostMEM.size() && !found ; h++) {
					Node otherhost = hostListMostMEM.get(h);
					if (taskCanFitOnNode(t, otherhost, jobRegistry.getRunningTasks(), migrateVM, migrateHost) && otherhost != leastLoadedNode && !Utility.nodeIsEmpty(otherhost, log)) {
//						log.debug("Task "+t.getJobId()+" can fit on host in MOSTMEMLIST "+otherhost.getFullName());
						//register migration
						migrateVM.add(t.getJobId());
						migrateHost.add(otherhost.getFullName());
						//update usage
//						otherhost.setCpuUsage(otherhost.getCpuUsage()+vm.getCpuNumber()*100);
//						otherhost.setMemUsage(otherhost.getMemUsage()+vm.getMem());
						//If needed change list of h
						if (getCPULoad(otherhost, jobRegistry.getRunningTasks(), migrateVM, migrateHost) < getMEMLoad(otherhost, jobRegistry.getRunningTasks(), migrateVM, migrateHost)) {
							hostListMostMEM.remove(otherhost);
							hostListMostCPU.add(otherhost);
						}
						found = true;
					}
				}
			}
			if (!found) {
//				log.debug("Could not find a suitable host for task : "+t.getJobId());
				success = false;
			}
		}
		//least loaded host is empty -> success = false
		if (migrateVM.isEmpty()) {
			success=false;
		}
		
		//if success, enforce realloc
		if (success) {
			for (int i = 0 ; i<migrateVM.size() ; i++) {
//				log.info("-- Migration -- : Task "+migrateVM.get(i)+" goes to node "+migrateHost.get(i));
			}
			log.info("Enforcing reallocation");
			
			//Write fmigrations into file (might be slow)
			PrintWriter writer = null;
			PrintWriter writer2 = null;
			try {
//				log.debug(System.getProperty("user.dir"));
				writer = new PrintWriter(new FileWriter(MIGRATION_PATH, true));
				writer2 = new PrintWriter(new FileWriter(ON_VS_EXPECTED, true));
				
//				log.debug("MigrateVM "+migrateVM.size()+" = "+migrateVM.get(0));
				for (int i = 0 ; i<migrateVM.size() ; i++) {
					ExecTask t = jobRegistry.getTask(migrateVM.get(i), "0");
					
//					log.debug("TaskID: "+t.getId()+"/"+t.getJobId());
					//should give the ok for migration 
					ArrayList<Node> nlist = new ArrayList<Node>();
					nlist.add((Node) resourceManager.getResourceByName(migrateHost.get(i)));
					
					log.debug("Destination host: "+nlist.get(0).getFullName());
					log.debug("Task : "+t.getJobId());
					Map<ResourceUnitName, ResourceUnit> destinationResources = chooseResourcesForExecution(nlist , t);
					jobRegistry.migrateTask(t.getJobId(), t.getId(), destinationResources);
					writer.println(String.format("%s\t%s", t.getJobId(), nlist.get(0).getName()));
					int onNodes=-1;
					for (Node n : resourceManager.getNodes()) {
						if (n.getPowerInterface().getPowerState().equals(StandardPowerStateName.ON)
								|| n.getPowerInterface().getPowerState().equals(StandardPowerStateName.BOOT))
							onNodes++;
					}
					//SI la migration peut amener une extinction
					if (Utility.getExpectedHostNumber(jobRegistry, resourceManager, log) < onNodes) 
						writer2.println("TRUE");
					else
						writer2.println("FALSE");
//					writer2.println(String.format("%d\t%d", Utility.getExpectedHostNumber(jobRegistry, resourceManager, log), onNodes));
				}
				writer.close();
				writer2.close();
			} catch (IOException e) {
				e.printStackTrace();
			} finally {
				writer.close();
			}
		}
	}

//	private void optimizeEnergyUsage(JobRegistry jobRegistry, ClusterResourceManager resourceManager, String resName) {
//		Node overLoadedNode = (Node)resourceManager .getResourceByName(resName);
//		Node leastLoadedNode = findLeastLoadedResource(resourceManager.getNodes());
//		
//		if(leastLoadedNode.getLoadInterface().getRecentUtilization().getValue() >= overLoadedNode.getLoadInterface().getRecentUtilization().getValue()){
//			return;
//		}
//
//		JobRegistry srcHostJr = new JobRegistryImpl(overLoadedNode.getFullName());
//		int runningTasksNumber = srcHostJr.getRunningTasks().size();
//		for(int i = srcHostJr.getRunningTasks().size() - 1; i >= 0 && runningTasksNumber > 2; i--) {
//			ExecTask execTask = srcHostJr.getRunningTasks().get(i);
//			Map<ResourceUnitName, ResourceUnit> destinationResources = chooseResourcesForExecution(findLeastLoadedResource(resourceManager.getNodes()).getProcessors(), execTask);
//			jobRegistry.migrateTask(execTask.getJobId(), execTask.getId(), destinationResources);
//			runningTasksNumber--;
//		}
//	}
	
	



	

	/**
	 * Is task CPUbound compared to reference
	 * @param t
	 * @return
	 */
	private boolean taskIsCPUBound(ExecTask t) {
		//ATTENTION ICI NB PROC EN DUR
		try {
			return ((float) t.getCpuCntRequest() / (float)400) > ((float)t.getMemoryRequest() / (float)16384.0);
		} catch (NoSuchFieldException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		log.debug("Could not determine if task "+t.getJobId()+" was cpubound");
		return true;
	}

	private float getMEMLoad(Node h, List<ExecTask> tasklist, ArrayList<String> migrateVM, ArrayList<String> migrateHost) {
		try {
			int additional = 0;
			if (tasklist != null && migrateVM != null && migrateHost != null) {
				//Get the migrating toward node h
				for (int i = 0 ; i < migrateHost.size() ; i++) {
					if (h.getFullName().equals(migrateHost.get(i))) {
						for(int j=0 ; j<tasklist.size() ; j++) {
							if (migrateVM.get(i).equals(tasklist.get(j))) {
								additional+=tasklist.get(j).getMemoryRequest();
							}
						}
					}
				}
			}
			
			return 1 - (((float)h.getFreeMemory()-(float)additional) / (float)h.getTotalMemory());
		} catch (NoSuchFieldException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		return 0;
	}

	private float getCPULoad(Node h, List<ExecTask> tasklist, ArrayList<String> migrateVM, ArrayList<String> migrateHost) {
		int additional = 0;
		if (tasklist != null && migrateVM != null && migrateHost != null) {
			//Get the migrating toward node h
			for (int i = 0 ; i < migrateHost.size() ; i++) {
				if (h.getFullName().equals(migrateHost.get(i))) {
					for(int j=0 ; j<tasklist.size() ; j++) {
						if (migrateVM.get(i).equals(tasklist.get(j))) {
							try {
								additional+=tasklist.get(j).getCpuCntRequest();
							} catch (NoSuchFieldException e) {
								// TODO Auto-generated catch block
								e.printStackTrace();
							}
						}
					}
				}
			}
		}
		return 1 - (((float) h.getFreeProcessorsNumber()-(float)additional)  / (float)h.getProcessorsNumber());
	}

	private void printNodes(List<Node> nodeList) {
		for (Node n : nodeList){
			try {
				log.debug(n.getName() + "("+n.getPowerInterface().getPowerState().getLabel()+"): " +
						"CPU="+n.getFreeProcessorsNumber()+"/"+n.getProcessorsNumber() +"(" + (float)n.getFreeProcessorsNumber() / (float)n.getProcessorsNumber()+ ")" +
						" / MEM: " +n.getFreeMemory()+"/"+n.getTotalMemory()+ "("+(float)n.getFreeMemory()/(float)n.getTotalMemory()+")");
			} catch (NoSuchFieldException e) {
				e.printStackTrace();
			}
		}
	}

	/**
	 * Gets the most loaded resource of the node n
	 * @param n
	 * @return
	 */
	private float getMostLoadedResource(Node n) {
		float memLoad = 0, cpuLoad;
		try {
			memLoad = (float)1 - (float)n.getFreeMemory() / (float)n.getTotalMemory();
		} catch (NoSuchFieldException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		cpuLoad = (float)1 - (float)n.getFreeProcessorsNumber() / (float)n.getProcessorsNumber();
//		log.debug("GetMostLoadedResource node "+n.getName()+": "+cpuLoad+" -- "+memLoad);
		if (cpuLoad > memLoad) {
			return cpuLoad;
		}
		else {
			return memLoad;
		}
	}
	
	/**
	 * Gets the least loaded host from the list
	 * Host powered on is not tested
	 * @param nodeList
	 * @return
	 */
	private Node getLeastLoadedHost(List<Node> nodeList) {
		Node n = null;
		float min = 2;
		for (int i = 0 ; i < nodeList.size() ; i++) {
			float mostLoadedResource = getMostLoadedResource(nodeList.get(i));
			if (min >= mostLoadedResource  && mostLoadedResource > 0) {
				n = nodeList.get(i);
				min = mostLoadedResource;
			}
		}
		if (n == null) {
			log.error("Least loaded node is null !");
			return null;
		}
//		logger.debug("Least loaded host is : "+h.getId() +" (CPU="+h.getCpuUsage()+"/MEM="+h.getMemUsage()+")");
		return n;
	}

	/**
	 * Dummy function to test full allocation to 1 proc
	 * @param processors
	 * @param task
	 * @return
	 */
	private Map<ResourceUnitName, ResourceUnit> chooseNode1(
			List<Processor> processors, TaskInterface<?> task) { 
		Map<ResourceUnitName, ResourceUnit> map = new HashMap<ResourceUnitName, ResourceUnit>(1);
		int cpuRequest = 0;
		try {
			cpuRequest = Double.valueOf(task.getCpuCntRequest()).intValue();
		} catch (NoSuchFieldException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		List<ComputingResource> choosenResources = new ArrayList<ComputingResource>(cpuRequest);
		choosenResources.add(processors.get(0));
		ProcessingElements pe = new ProcessingElements();
		pe.addAll(choosenResources);
		map.put(StandardResourceUnitName.PE, pe);
		return map;
	}
	
	/**
	 * 
	 * @param processors
	 * @param task
	 * @return
	 */
	private Map<ResourceUnitName, ResourceUnit> chooseResourcesForExecution(
			List<Node> nodes, TaskInterface<?> task) {

		Map<ResourceUnitName, ResourceUnit> map = new HashMap<ResourceUnitName, ResourceUnit>(1);
		
		List<ComputingResource> chosenResources = new ArrayList<ComputingResource>();	
		boolean found = false;
		Memory memory = null;
		int cpuRequest=0;
		
		Collections.sort(nodes, new ComparatorHostHighestLoadFirst());
		
		for (int h=0 ; h<nodes.size() && !found ; h++){
			Node n = nodes.get(h);
//			log.debug("Task "+task.getId()+"/"+task.getJobId()+" can fit?");
			if (taskCanFitOnNode(task, n, null, null, null)){
				try {
					log.debug("Task "+task.getId()+"/"+task.getJobId()+" (CPU="+task.getCpuCntRequest()+" ; MEM="+task.getMemoryRequest()+") can fit on node "+n.getFullName());
				} catch (NoSuchFieldException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
				found=true;
				
				try {
					cpuRequest = (int)task.getCpuCntRequest();
					memory = new Memory(n.getMemory(), (int)task.getMemoryRequest(), (int)task.getMemoryRequest());
				} catch (NoSuchFieldException e1) {
					e1.printStackTrace();
				}
				//Adding chosen processors
				for (int i = 0; i < n.getProcessors().size() && cpuRequest > 0; i++) {
					if (n.getProcessors().get(i).getStatus() == ResourceStatus.FREE) {
						chosenResources.add(n.getProcessors().get(i));
						cpuRequest--;
					}
				}
				if (cpuRequest > 0) {
					continue;
				}
			}
		}
		if (!found) {
			log.info("Allocation not found for task "+task.getJobId());
			return null;
		}
		
		ProcessingElements pe = new ProcessingElements();
		pe.addAll(chosenResources);
		map.put(StandardResourceUnitName.PE, pe);
		if (memory != null)
			map.put(StandardResourceUnitName.MEMORY, memory);
		else 
			return null;
		return map;
	}
	
//	/**
//	 * FCFS CPU
//	 * @param processors
//	 * @param task
//	 * @return
//	 */
//	private Map<ResourceUnitName, ResourceUnit> chooseResourcesForExecution(
//			List<Processor> processors, TaskInterface<?> task) {
//
//		Map<ResourceUnitName, ResourceUnit> map = new HashMap<ResourceUnitName, ResourceUnit>(1);
//
//		int cpuRequest;
//		try {
//			cpuRequest = Double.valueOf(task.getCpuCntRequest()).intValue();
//		} catch (NoSuchFieldException e) {
//			cpuRequest = 0;
//		}
//
//		if (cpuRequest != 0) {
//
//			if (processors.size() < cpuRequest) {
//				return null;
//			}
//
//			List<ComputingResource> choosenResources = new ArrayList<ComputingResource>(cpuRequest);				
//			for (int i = 0; i < processors.size() && cpuRequest > 0; i++) {
//				if (processors.get(i).getStatus() == ResourceStatus.FREE) {
//					choosenResources.add(processors.get(i));
//					cpuRequest--;
//				}
//			}
//			if (cpuRequest > 0) {
//				return null;
//			}
//
//			ProcessingElements pe = new ProcessingElements();
//			pe.addAll(choosenResources);
//			map.put(StandardResourceUnitName.PE, pe);
//			return map;
//		}
//
//		return null;
//	}

	/**
	 * Checks whether a task can fit on a node
	 * @param task
	 * @param n
	 * @param migrateVM 
	 * @param runningTasks 
	 * @param migrateHost 
	 * @return
	 */
	private boolean taskCanFitOnNode(TaskInterface<?> task, Node n, List<ExecTask> runningTasks, ArrayList<String> migrateVM, ArrayList<String> migrateHost) {
		int cpuRequest;
		try {
			cpuRequest = Double.valueOf(task.getCpuCntRequest()).intValue();
		} catch (NoSuchFieldException e) {
			log.warn("Task : "+task.getId()+" does not require CPU !");
			cpuRequest = 0;
		}
		
		int memRequest;
		try {
			memRequest = (int) task.getMemoryRequest();
		} catch (NoSuchFieldException e) {
			log.warn("Task : "+task.getJobId()+" does not require MEMORY !");
			memRequest = 0;
		}
//		log.debug("Task "+task.getId()+" => CPU="+cpuRequest+"/MEM="+memRequest);

		if (cpuRequest == 0 && memRequest == 0){
			log.error("Error: task is empty !");
			return false;
		}
		
		try {
			
			//To for expected migrations
			int additionalProc = 0;
			int additionalMem = 0;
			if (runningTasks != null && migrateVM != null && migrateHost != null) {
				//check node is destination
				for (int i = 0 ; i < migrateHost.size() ; i++) {
					String nodeName = migrateHost.get(i);
					if (n.getFullName().equals(nodeName)) {
						String migrateTaskId = migrateVM.get(i);
						//get corresponding task
						for (int j = 0 ; j<runningTasks.size() ; j++) {
							if (runningTasks.get(j).getJobId().equals(migrateTaskId)) {
								additionalProc+=runningTasks.get(j).getCpuCntRequest();
								additionalMem+=runningTasks.get(j).getMemoryRequest();
							}
						}
					}
				}
				
			}
			//For each processor check whether it can fit
			if (n.getFreeMemory() >= memRequest+additionalMem && n.getFreeProcessorsNumber() >= cpuRequest+additionalProc) {
				return true;
			}
			
		} catch (NoSuchFieldException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		return false;
	}

	private Node findLeastLoadedResource(List<Node> nodes ){
		Node leastLoadedNode = null;
		double minLoad = Double.MAX_VALUE;
		
		for(int i = 0; i < nodes.size(); i++){
			Node node = nodes.get(i);
			double totalLoad = node.getLoadInterface().getRecentUtilization().getValue();
			if(totalLoad < minLoad){
				leastLoadedNode= node;
				minLoad = totalLoad;
			}
		}
		return leastLoadedNode;
	}

	private void sendBackTask(TaskInterface<?> task, JobRegistry jobRegistry, TaskQueue q, String clusterName) {
		try {
			PrintWriter writer = new PrintWriter(new FileWriter(TASK_REJECTIONS, true));
			//Task id / Cluster name 
			writer.write(String.format("%s\t%s\n", task.getJobId(), clusterName));
			writer.close();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		log.debug("Task: "+task.getJobId() + " sent back to GridBroker");
		try {
			//Unsubmit task
			jobRegistry.getJobInfo(task.getJobId()).getTask(task.getId()).setStatus((int)BrokerConstants.TASK_STATUS_UNSUBMITTED);

		} catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		q.remove(task);
		synchronized(jobRegistry) {
			jobRegistry.getTasks().remove(task);
		}
		DataCenterWorkloadSimulator.getEventManager().send("grid", 0, GridSimTags.GRIDLET_SUBMIT, jobRegistry.getJobInfo(task.getJobId()));
		
	}
}
